#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mpi.h"

int main( int argc, char *argv[] )
{   
    int myrank, nprocs;
    int n, vec_size, mat_size;
    int i, j;
    float *sm1, // SM: submatrix
          *vec;
    MPI_Request request;

    MPI_Init(&argc,&argv);

    if ( argc != 2 )
    {
        fprintf( stderr, "Usage: %s n\n", argv[0] );
        exit( -1 );
    }

    n = atoi(argv[1]);
    mat_size = (n/2) * (n/2);
    vec_size = n/2;

    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );

    sm1 = (float *) malloc ( mat_size * sizeof(float) );
    vec = (float *) malloc ( vec_size * sizeof(float) );

    // Initialize my submatrix of A (sm1)
    for ( i = 0; i < mat_size; i++ )
        sm1[ i ] = myrank;

    // Communication step 1 ( row-wise )
    if ( myrank % 2 == 0 ) // I'm column 0
    {
        for ( i = 0; i < n/2; i++ )
        {
            MPI_Isend( &sm1[i*vec_size], vec_size, MPI_FLOAT, 
                      myrank+1, 0, MPI_COMM_WORLD, &request ); // Send my vector
            MPI_Recv( vec, vec_size, MPI_FLOAT, 
                      myrank+1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv vector
            MPI_Wait( &request, MPI_STATUS_IGNORE );
            // Computation
            for ( j = 0; j < vec_size; j++ ) // Column 0: addition
                sm1[i*vec_size+j] = sm1[i*vec_size+j] + vec[j]; // sm1: B
        }
    }
    else // I'm column 1
    {
        for ( i = 0; i < n/2; i++ )
        {
            MPI_Isend( &sm1[i*vec_size], vec_size, MPI_FLOAT, 
                      myrank-1, 0, MPI_COMM_WORLD, &request ); // Send my vector
            MPI_Recv( vec, vec_size, MPI_FLOAT, 
                      myrank-1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv vector
            MPI_Wait( &request, MPI_STATUS_IGNORE );
            // Computation
            for ( j = 0; j < vec_size; j++ ) // Column 1: substraction
                sm1[i*vec_size+j] = vec[j] - sm1[i*vec_size+j]; // sm1: B
        }
    }

    // Communication step 2
    if ( myrank / 2 == 0 ) // I'm row 0
    {
        for ( i = 0; i < n/2; i++ )
        {
            MPI_Isend( &sm1[i*vec_size], vec_size, MPI_FLOAT, 
                      myrank+2, 1, MPI_COMM_WORLD, &request ); // Send my vector
            MPI_Recv( vec, vec_size, MPI_FLOAT, 
                      myrank+2, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv vector
            MPI_Wait( &request, MPI_STATUS_IGNORE );
            // Computation
            for ( j = 0; j < vec_size; j++ ) // Row 0: addition
                sm1[i*vec_size+j] = sm1[i*vec_size+j] + vec[j]; // sm1: C
        }
    }
    else // I'm row 1
    {
        for ( i = 0; i < n/2; i++ )
        {
            MPI_Isend( &sm1[i*vec_size], vec_size, MPI_FLOAT, 
                      myrank-2, 1, MPI_COMM_WORLD, &request ); // Send my vector
            MPI_Recv( vec, vec_size, MPI_FLOAT, 
                      myrank-2, 1, MPI_COMM_WORLD, MPI_STATUS_IGNORE ); // Recv vector
            MPI_Wait( &request, MPI_STATUS_IGNORE );
            // Computation
            for ( j = 0; j < vec_size; j++ ) // Row 1: substraction
                sm1[i*vec_size+j] = vec[j] - sm1[i*vec_size+j]; // sm1: C
        }
    }

    // Check result
    float res[4] = { 6, -2, -4, 0};
    int myres = res[myrank];
    int error = 0;
    for ( i = 0; i < mat_size; i++ )
        sm1[i] -= myres;
    for ( i = 0; i < mat_size; i++ )
        if ( fabs( sm1[i] ) > 1e-5 ) // some threshold
        {
            fprintf( stderr, "[%d] %d: %.7f\n", myrank, i, sm1[i] );
            error = 1;
            break;
        }

    if ( error )
        printf("[%d] ERROR!\n", myrank);
    else
        printf("[%d] WORKED!\n", myrank);

    free( sm1 );
    free( vec );

    MPI_Finalize();

    return 0;
}
